# Globals.


class style:
    """
    Styling global options for misura.
    """

    quantityHighlighting = True
    quantityPlusMinus = " \u00b1 "


class defined:
    """
    User defined units.
    """

    BASE_TABLE = {}
    DERIVED_TABLE = {}
    DERIVED_UNPACKING_TABLE = {}
